/*-------------------------------------------------------------------------
 Copyright 2006 Olivier Berlanger

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 -------------------------------------------------------------------------*/
package net.sf.xolite.impl;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import net.sf.xolite.XMLSerializeException;
import net.sf.xolite.XMLSerializer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * Base for XMLSerializer implementations. This class hold all the complex state of XML serialization (like mapped prefixes,
 * elements currently started, ...). It verify the coherence of the state (like checking that the ended element is the element
 * currently started) and generate simpler serialization events (the abstract methods). Those new events will be implemented by
 * concrete sub-classes.
 */
public abstract class BaseXMLSerializer implements XMLSerializer {


    private static Log log = LogFactory.getLog(BaseXMLSerializer.class);

    protected static final int STATE_INIT = 0;
    protected static final int STATE_ELEMENT_START_OPENED = 1;
    protected static final int STATE_ELEMENT_START_CLOSED = 2;
    protected static final int STATE_TEXT = 3;
    protected static final int STATE_TEXT_MULTI_LINE = 4;
    protected static final int STATE_ELEMENT_CLOSED = 5;
    protected static final int STATE_FINISHED = -1;

    private int state = STATE_INIT;
    private int level = -1;
    private List<LevelState> levelStates;
    private int attributeIndex;
    private Map<Object, Object> customObjects;


    protected BaseXMLSerializer() {
        levelStates = new ArrayList<LevelState>();
    }


    protected int getLevel() {
        return level;
    }


    public void startDocument() throws XMLSerializeException {
        state = STATE_INIT;
        level = -1;
        try {
            startDocumentImpl();
        } catch (XMLSerializeException xse) {
            throw xse;
        } catch (Exception e) {
            throw new XMLSerializeException("Cannot start document", e);
        }
    }


    public void endDocument() throws XMLSerializeException {
        if (level >= 0) throw new XMLSerializeException("All tags are not closed at the end of the document");
        try {
            endDocumentImpl();
        } catch (XMLSerializeException xse) {
            throw xse;
        } catch (Exception e) {
            throw new XMLSerializeException("Cannot end document", e);
        }
        state = STATE_FINISHED;
    }


    private LevelState increaseLevel() throws XMLSerializeException {
        if ((level < 0) && (state != STATE_INIT))
            throw new XMLSerializeException("Bad documents state: " + state + ", should be STATE_INIT");
        LevelState levelSt;
        level++;
        int len = levelStates.size();
        if (level < len) levelSt = levelStates.get(level);
        else if (level == len) {
            LevelState parent = (level == 0) ? null : (LevelState) levelStates.get(level - 1);
            levelSt = new LevelState(parent);
            levelStates.add(levelSt);
        } else {
            throw new IllegalStateException("Invalid level=" + level + " when levelStates.size()=" + len);
        }
        return levelSt;
    }


    private void decreaseLevel() throws XMLSerializeException {
        if (level < 0) throw new XMLSerializeException("No more XML element open");
        LevelState levelSt = levelStates.get(level);
        levelSt.clear();
        level--;
    }


    private LevelState getLevelState() {
        return levelStates.get(level);
    }


    private LevelState getNextLevelState() throws XMLSerializeException {
        LevelState levelSt = increaseLevel();
        level--;
        return levelSt;
    }


    /**
     * Get the element name of the last opened level (to be used for debug output only).
     */
    protected String getLastElementName() {
        return getLevelState().getCurrentTag();
    }


    /**
     * Get the tag length (prefix+name) of the last opened level.
     */
    protected int getLastElementLength() {
        LevelState lvlState = getLevelState();
        int len = lvlState.getCurrentTag().length();
        String prefix = lvlState.getCurrentPrefix();
        if ((prefix != null) && !prefix.equals("")) {
            len += prefix.length() + 1;
        }
        return len;
    }


    private void closeElementStart(boolean complexContent) throws XMLSerializeException {
        if (state == STATE_ELEMENT_START_OPENED) {
            try {
                closeElementStartImpl(complexContent);
            } catch (XMLSerializeException xse) {
                throw xse;
            } catch (Exception e) {
                throw new XMLSerializeException("Cannot finalize element start", e);
            }
            state = STATE_ELEMENT_START_CLOSED;
        }
    }


    // ------------------------ XMLSerializer interface implementation -------------------------


    public void startPrefixMapping(String prefix, String namespaceUri) throws XMLSerializeException {
        if (prefix == null)
            throw new XMLSerializeException("Cannot define mapping for null prefix, use \"\" to define default prefix");
        if ((namespaceUri == null) || "".equals(namespaceUri))
            throw new XMLSerializeException("Cannot define mapping for null namespaceUri");

        if (state == STATE_ELEMENT_START_OPENED) {
            // define the prefix in this level (and possibly add definition directly)
            try {
                LevelState levelSt = getLevelState();
                if (levelSt.mapPrefix(prefix, namespaceUri)) definePrefixMapping(prefix, namespaceUri);
            } catch (XMLSerializeException xse) {
                throw xse;
            } catch (Exception e) {
                throw new XMLSerializeException("Cannot define prefix mapping " + prefix + "=" + namespaceUri, e);
            }
        } else {
            // define the prefix for the next level
            LevelState levelSt = getNextLevelState();
            levelSt.mapPrefix(prefix, namespaceUri);
        }
    }


    public void startElement(String uri, String tag) throws XMLSerializeException {
        if (tag == null) throw new XMLSerializeException("Element Tag cannot be null");
        if (tag.length() == 0) throw new XMLSerializeException("Element Tag cannot be empty string");
        closeElementStart(true);
        LevelState levelSt = increaseLevel();
        if (log.isDebugEnabled()) log.debug("Start element <" + tag + "> level = " + level);
        if (uri == null) uri = levelSt.getDefaultNamespaceUri();
        String prefix = levelSt.getPrefix(uri);
        levelSt.setCurrentValues(uri, prefix, tag);
        attributeIndex = 0;
        try {
            startElementImpl(uri, prefix, tag);
            // write all namespace definitions of this level
            for (Iterator<String> iter = levelSt.getNamespaceDefinitons(); iter.hasNext();) {
                String defUri = iter.next();
                String defPrefix = levelSt.getPrefix(defUri);
                definePrefixMapping(defPrefix, defUri);
            }
        } catch (XMLSerializeException xse) {
            throw xse;
        } catch (Exception e) {
            throw new XMLSerializeException("Cannot start element", e);
        }
        state = STATE_ELEMENT_START_OPENED;
    }


    public void attribute(String name, String value) throws XMLSerializeException {
        attribute(null, name, value);
    }


    public void attribute(String uri, String name, String value) throws XMLSerializeException {
        if (state != STATE_ELEMENT_START_OPENED)
            throw new XMLSerializeException("writeAttribute must be called just after writeElementStart");
        if (value == null) value = "";
        LevelState levelSt = getLevelState();
        String prefix = levelSt.getPrefix(uri);
        try {
            attributeImpl(uri, prefix, name, value, attributeIndex);
            attributeIndex++;
        } catch (XMLSerializeException xse) {
            throw xse;
        } catch (Exception e) {
            throw new XMLSerializeException("Cannot serialize attribute", e);
        }
    }


    public void characters(String text) throws XMLSerializeException {
        if ((text != null) && (!text.equals(""))) {
            closeElementStart(false);
            try {
                charactersImpl(text, false);
            } catch (XMLSerializeException xse) {
                throw xse;
            } catch (Exception e) {
                throw new XMLSerializeException("Cannot serialize text", e);
            }
            state = STATE_TEXT;
        }
    }


    public void charactersMultiLine(String text) throws XMLSerializeException {
        if ((text != null) && (!text.equals(""))) {
            closeElementStart(false);
            try {
                charactersImpl(text, true);
            } catch (XMLSerializeException xse) {
                throw xse;
            } catch (Exception e) {
                throw new XMLSerializeException("Cannot serialize text", e);
            }
            state = STATE_TEXT_MULTI_LINE;
        }
    }


    public void endElement(String uri, String tag) throws XMLSerializeException {
        LevelState levelSt = getLevelState();
        levelSt.checkCurrentValues(uri, tag);
        String prefix = levelSt.getPrefix(uri);
        try {
            if (state == STATE_ELEMENT_START_OPENED) {
                endInlineElementImpl(uri, prefix, tag);
            } else if ((state == STATE_ELEMENT_START_CLOSED) || (state == STATE_TEXT) || (state == STATE_TEXT_MULTI_LINE)) {
                endTextElementImpl(uri, prefix, tag, state == STATE_TEXT_MULTI_LINE);
            } else {
                endComplexElementImpl(uri, prefix, tag);
            }
        } catch (XMLSerializeException xse) {
            throw xse;
        } catch (Exception e) {
            throw new XMLSerializeException("Cannot end element", e);
        }
        decreaseLevel();
        if (log.isDebugEnabled()) log.debug("End element <" + tag + "> leval = " + level);
        state = STATE_ELEMENT_CLOSED;
    }


    public void simpleElement(String namespaceUri, String localName, String text) throws XMLSerializeException {
        startElement(namespaceUri, localName);
        characters(text);
        endElement(namespaceUri, localName);
    }


    // ----------------------- Abstract methods ----------------------


    protected abstract void startDocumentImpl() throws Exception;


    protected abstract void endDocumentImpl() throws Exception;


    protected abstract void startElementImpl(String uri, String prefix, String tag) throws Exception;


    protected abstract void attributeImpl(String uri, String prefix, String name, String value, int index) throws Exception;


    protected abstract void closeElementStartImpl(boolean complexContent) throws Exception;


    protected abstract void charactersImpl(String text, boolean multilineContent) throws Exception;


    protected abstract void endInlineElementImpl(String uri, String prefix, String tag) throws Exception;


    protected abstract void endTextElementImpl(String uri, String prefix, String tag, boolean multilineContent)
            throws Exception;


    protected abstract void endComplexElementImpl(String uri, String prefix, String tag) throws Exception;


    protected void definePrefixMapping(String prefix, String uri) throws Exception {
        String xmlns = "xmlns";
        if ("".equals(prefix)) {
            prefix = "xmlns";
            xmlns = null;
        }
        attributeImpl(uri, xmlns, prefix, uri, attributeIndex);
        attributeIndex++;
    }


    public Object getCustomObject(Object key) {
        return (customObjects == null) ? null : customObjects.get(key);
    }


    public void putCustomObject(Object key, Object value) {
        if (customObjects == null) customObjects = new HashMap<Object, Object>();
        customObjects.put(key, value);
    }


    // ----------------------- Inner class maintaining XML level state information ----------------------


    /**
     * Class holding the current state of a level in the path to the current element.
     */
    private static class LevelState {


        private String defaultNamespaceUri;
        private String currentNamespaceUri;
        private String currentPrefix;
        private String currentTag;
        private Map<String, String> prefixMappings;
        private LevelState parent;


        LevelState(LevelState parentState) {
            parent = parentState;
            prefixMappings = new HashMap<String, String>();
        }


        String getCurrentPrefix() {
            return currentPrefix;
        }


        String getCurrentTag() {
            return currentTag;
        }


        void clear() {
            defaultNamespaceUri = null;
            currentNamespaceUri = null;
            currentPrefix = null;
            currentTag = null;
            prefixMappings.clear();
        }


        void setCurrentValues(String uri, String prefix, String tag) {
            currentNamespaceUri = uri;
            currentPrefix = prefix;
            currentTag = tag;
        }


        void checkCurrentValues(String uri, String tag) throws XMLSerializeException {
            if (!checkEquals(currentNamespaceUri, uri))
                throw new XMLSerializeException("The closed element namespace URI is not the same as the opened one: "
                        + currentNamespaceUri + " != " + uri);
            if (!checkEquals(currentTag, tag))
                throw new XMLSerializeException("The closed element tag is not the same as the opened one: " + currentTag
                        + " != " + tag);
        }


        private boolean checkEquals(String str1, String str2) {
            if (str1 == str2) return true;
            if (str1 == null) return str2 != null;
            return str1.equals(str2);
        }


        String getPrefix(String namespaceUri) throws XMLSerializeException {
            String prefix = null;
            if (namespaceUri != null) {
                prefix = searchPrefix(namespaceUri);
                if (prefix == null) {
                    char nsChar = 'a';
                    while (prefix == null) {
                        String candidatePrefix = "" + nsChar + nsChar;
                        String uri = searchNamespaceUri(candidatePrefix);
                        if (uri == null) {
                            prefix = candidatePrefix;
                        }
                    }
                    mapPrefix(prefix, namespaceUri);
                }
            }
            return prefix;
        }


        boolean mapPrefix(String prefix, String namespaceUri) {
            boolean added = false;
            String existingPrefix = searchPrefix(namespaceUri);
            if (existingPrefix == null) {
                prefixMappings.put(namespaceUri, prefix);
                if (prefix.equals("")) defaultNamespaceUri = namespaceUri;
                added = true;
            }
            return added;
        }


        private String searchPrefix(String namespaceUri) {
            String prefix = null;
            if (checkEquals(currentNamespaceUri, namespaceUri)) prefix = currentPrefix;
            if (prefix == null) {
                prefix = prefixMappings.get(namespaceUri);
                if ((prefix == null) && (parent != null)) prefix = parent.searchPrefix(namespaceUri);
            }
            return prefix;
        }


        private String searchNamespaceUri(String prefix) {
            String namespaceUri = null;
            if (checkEquals(currentPrefix, prefix)) namespaceUri = currentNamespaceUri;
            if (namespaceUri == null) {
                for (String ns : prefixMappings.keySet()) {
                    if (prefix.equals(prefixMappings.get(ns))) {
                        namespaceUri = ns;
                        break;
                    }
                }
                if ((namespaceUri == null) && (parent != null)) namespaceUri = parent.searchNamespaceUri(prefix);
            }
            return namespaceUri;
        }


        Iterator<String> getNamespaceDefinitons() {
            return prefixMappings.keySet().iterator();
        }


        String getDefaultNamespaceUri() {
            if ((defaultNamespaceUri == null) && (parent != null)) defaultNamespaceUri = parent.getDefaultNamespaceUri();
            return defaultNamespaceUri;
        }

    }

}
