package net.sf.xolite.utils;


import java.net.URL;


/**
 * Helper to build the the schema location expressed as in the standard 'xsi:schemaLocation' xml attributes. <br>
 * Such a string can be set on the SaxXMLEventParser to enable validation.
 * <p>
 * <b>Caution: </b> This schemaLocation attribute setup works with Sun JRE default parser (Xerces) but is not guaranteed to work
 * if you (or a library you're using) have changed the default XML parser implementation. <br>
 * In this case, you have to setup a validating parser yourself (following your parser documentation) and pass it to the
 * <code>SaxXMLEventParser(XMLReader)</code> constructor.
 * </p>
 * 
 * @author berol
 */
public class SchemaLocation {


    /**
     * Get a schemaLocation string for a schema packaged as an application resource.
     * 
     * @param schemaUri
     *            URI of the schema
     * @param referenceClass
     *            the reference class providing the class loader.
     * @param resourcePath
     *            Path to the resource, it can be absolute (i.e. starting with '/') or relative to the package of this reference
     *            class.
     * @return the schemaLocation string for the specified schema.
     */
    public static String getResourceLocation(String schemaUri, Class<?> referenceClass, String resourcePath) {
        return schemaUri + " " + getResourceURL(referenceClass, resourcePath);
    }


    /**
     * Return the URL to a resource. <br>
     * This method verifies the existence of the requested resource and throw a IllegalArgumentException if the ressource is not
     * found.
     * 
     * @param referenceClass
     *            the reference class providing the class loader.
     * @param resourcePath
     *            Path to the resource, it can be absolute (i.e. starting with '/') or relative to the package of this reference
     *            class.
     * @return the URL to the specified resource.
     * @exception IllegalArgumentException
     *                if the specified resource is not found.
     */
    public static String getResourceURL(Class<?> referenceClass, String resourcePath) {
        URL resourceUrl = referenceClass.getResource(resourcePath);
        if (resourceUrl == null) {
            String fullPath = resourcePath;
            if (!resourcePath.startsWith("/")) {
                fullPath = referenceClass.getPackage().getName().replace('.', '/') + "/" + resourcePath;
            }
            throw new IllegalArgumentException("Cannot find resource in classpath: " + fullPath);
        }
        return resourceUrl.toString().replace(" ", "%20");
    }


}
